-----------------------------------------------------------------------------

local TagCompiler = AceLibrary("TagCompiler-1.0")

-----------------------------------------------------------------------------

Assessment.TagData = 
{ 
	TagError = function(x) Assessment:Print(x) end, 
--	TagDebug = function(x) Assessment:Debug(x) end 
}

-----------------------------------------------------------------------------

for k,v in pairs(TagCompiler.CommonTagData) do
	Assessment.TagData[k] = v
end

-----------------------------------------------------------------------------

Assessment.TagData.amount =
{
	tableData = "value.amount",
	numeric = true
}

Assessment.TagData.barvalue =
{
	tableData = "barValue",
	guaranteeResult = true,
	numeric = true
}

Assessment.TagData.count =
{
	tableData = "value.count",
	numeric = true
}

Assessment.TagData.formattime =
{
	method = function(data, value)
				if type(value) == "number" then
					return date("%H:%M:%S", value)
				end
				return value
			 end,
	noGuaranteeChange = true
}

Assessment.TagData.formatduration =
{
	method = function(data, value)
				if type(value) == "number" then
					if value < 60 then
						return ("%.1f"):format(value)
					else
						return ("%d:%02d.%d"):format(value/60, value%60, value%1*10)
					end
				end
				return value
			 end,
	noGuaranteeChange = true
}

Assessment.TagData.isplayerincombat =
{
	compileString = function() return "UnitAffectingCombat(\"player\")" end
}

Assessment.TagData.key =
{
	tableData = "key",
}

Assessment.TagData.maximumbarvalue =
{
	tableData = "maximumBarValue",
	guaranteeResult = true,
	numeric = true
}

Assessment.TagData.rank =
{
	tableData = "rank",
	guaranteeResult = true,
	numeric = true
}

Assessment.TagData.totalbarvalue =
{
	tableData = "totalBarValue",
	guaranteeResult = true,
	numeric = true
}

Assessment.TagData.value =
{
	tableData = "value",
}

-----------------------------------------------------------------------------

local compileCache =
{
	["string"] = { data = { }, events = { }, tag = "", method = function() return "" end },
}

function Assessment:CompileTag(tag, resultType)
	local cacheTag = resultType..tag
	if compileCache[cacheTag] then
		return compileCache[cacheTag]
	end

	local result = TagCompiler:Compile(tag, self.TagData, resultType)
	compileCache[cacheTag] = result
	return result
end

-----------------------------------------------------------------------------
