local MAJOR_VERSION = "Threat-1.0"
local MINOR_VERSION = tonumber(("$Revision: 38731 $"):match("%d+"))

if MINOR_VERSION > _G.ThreatLib_MINOR_VERSION then
	_G.ThreatLib_MINOR_VERSION = MINOR_VERSION
end

ThreatLib_funcs[#ThreatLib_funcs+1] = function()

useParser30 = AceLibrary:HasInstance("Parser-3.0")

local new, del, newHash, newSet = ThreatLib.new, ThreatLib.del, ThreatLib.newHash, ThreatLib.newSet

local ThreatLib = _G.ThreatLib
local ThreatLibNPCModuleCore = nil
if useParser30 then
	ThreatLibNPCModuleCore = ThreatLib:NewModule("NPCCore", "AceEvent-2.0", "Parser-3.0")
else
	ThreatLibNPCModuleCore = ThreatLib:NewModule("NPCCore", "AceEvent-2.0")
end

local L = ThreatLib.L

local modules = {}
local activeModules = {}

ThreatLibNPCModuleCore.modulePrototype = {}
function ThreatLibNPCModuleCore:GetModule(name)
	return modules[name]
end
local mt = {__index = ThreatLibNPCModuleCore.modulePrototype}
function ThreatLibNPCModuleCore:NewModule(name, ...)
	local mod = setmetatable({ mixins = {} }, mt)
	for i = 1, select('#', ...) do
		local lib = AceLibrary((select(i, ...)))
		mod.mixins[lib] = true
		lib:embed(mod)
	end
	local AceEvent = AceLibrary("AceEvent-2.0")
	if not mod.mixins[AceEvent] then
		mod.mixins[AceEvent] = true
		AceEvent:embed(mod)
	end
	if useParser30 then
		local Parser = AceLibrary("Parser-3.0")
		if not mod.mixins[Parser] then
			mod.mixins[Parser] = true
			Parser:embed(mod)
		end
	end
	modules[name] = mod
	mod.name = name
	activeModules[mod] = false
	return mod
end
function ThreatLibNPCModuleCore:ToggleModuleActive(module, value)
	if type(module) == "string" then
		module = modules[module]
	end
	if value == activeModules[module] then
		return
	end

	activeModules[module] = value
	
	local OnEmbedChange = value and "OnEmbedEnable" or "OnEmbedDisable"
	for mixin in pairs(module.mixins) do
		if type(mixin[OnEmbedChange]) == "function" then
			local success, ret = pcall(mixin[OnEmbedChange], mixin, module)
			if not success then
				geterrorhandler()(ret)
			end
		end
	end
	
	local OnChange = value and "OnEnable" or "OnDisable"
	if type(module[OnChange]) == "function" then
		local success, ret = pcall(module[OnChange], module)
		if not success then
			geterrorhandler()(ret)
		end
	end
end

local combatantModuleLookup = {}

local function ModifyThreat(mob, target, multi, add)
	local module
	if target == "player" then
		module = ThreatLib:GetPlayerModule()
	elseif target == "pet" and ThreatLib:DoesPlayerHavePet() then
		module = ThreatLib:GetPetModule()
	end
	if module then
		module:_multiplyTargetThreat(multi, mob)
		module:_addTargetThreat(add, mob)
	end
end

function ThreatLibNPCModuleCore:OnInitialize()
	self.activeModule = nil
	self.generic_attacks = {}
	self.generic_attacks[L["Knock Away"]] = function(mob, target) ModifyThreat(mob, target, 0.5, 0) end
	self.generic_attacks[L["Wing Buffet"]] = function(mob, target) ModifyThreat(mob, target, 0.5, 0) end
	self.generic_attacks[L["Sand Blast"]] = function(mob, target) ModifyThreat(mob, target, 0, 0) end		
	if useParser30 then
		self:RegisterParserEvent({
			eventType = "Damage",
			recipientID = "player"
		}, "parseMobAttack")
		self:RegisterParserEvent({
			eventType = "Damage",
			recipientID = "pet"
		}, "parseMobAttack")	
		self:RegisterParserEvent({
			eventType = "Death"
		}, "parseDeath")
	else
		-- TODO: Parser-1.1 compat
	end
	
	self.encounterMobCount = 1
	self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
	
	ThreatLib:ToggleModuleActive(self, true)
	
	for name, mod in pairs(modules) do
		if mod.OnInitialize then
			mod:OnInitialize()
		end
	end
end

function ThreatLibNPCModuleCore:parseMobAttack(info)
	if self.activeModule and self.activeModule:parseMobAttack(info) then
		return
	end

	if self.generic_attacks[info.abilityName] then
		self.generic_attacks[info.abilityName](info.sourceName, info.recipientID)
	end	
end

function ThreatLibNPCModuleCore:parseDeath(info)
	if self.activeModule then
		self.activeModule:parseDeath(info)
	end
end

function ThreatLibNPCModuleCore:UPDATE_MOUSEOVER_UNIT()
	if not UnitExists("mouseover") or UnitIsDead("mouseover") then
		return
	end
	local mobName = UnitName("mouseover")
	if combatantModuleLookup[mobName] and self.activeModule ~= combatantModuleLookup[mobName] then
		ThreatLibNPCModuleCore:ToggleModuleActive(combatantModuleLookup[mobName], true)
	end
end

function ThreatLibNPCModuleCore.modulePrototype:OnInitialize()
	self.chatEvents = {}
	self.encounterEnemies = {}
	self.attacks = {}
	self.encounterMobCount = 1
	self:Init()
	self.ModifyThreat = ModifyThreat
end

function ThreatLibNPCModuleCore.modulePrototype:OnEnable()
	self:RegisterEvent("CHAT_MSG_MONSTER_EMOTE")
	self:RegisterEvent("CHAT_MSG_MONSTER_YELL")
	
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	
	-- Replace any active module with ourselves after a clean shutdown
	if ThreatLibNPCModuleCore.activeModule then
		ThreatLibNPCModuleCore:ToggleModuleActive(ThreatLibNPCModuleCore.activeModule, false)
	end
	ThreatLibNPCModuleCore.activeModule = self
end

function ThreatLibNPCModuleCore.modulePrototype:OnDisable()
	ThreatLibNPCModuleCore.activeModule = nil
end

function ThreatLibNPCModuleCore.modulePrototype:parseDeath(info)
	if not ThreatLib.running then
		return
	end
	if not ThreatLib.inCombat then
		return
	end
	local recipientName = info.recipientName
	local func = self.encounterEnemies[recipientName]
	if func then
		func(self, recipientName, info.sourceName)
		self:TriggerEvent("ThreatLib_BossDeath", recipientName)
	end
end

function ThreatLibNPCModuleCore.modulePrototype:parseMobAttack(info)
	if not ThreatLib.running then
		return
	end
	if not ThreatLib.inCombat then
		return
	end
	local func = self.attacks[info.abilityName]
	if func then
		func(self, info.sourceName, info.recipientID)
		return true
	end
end

function ThreatLibNPCModuleCore.modulePrototype:OnDisable()
	--[[
	for evtType, mobs in pairs(self.chatEvents) do
		for mob, texts in pairs(mobs) do
			self.chatEvents[evtType][mob] = del(self.chatEvents[evtType][mob])
		end
		self.chatEvents[evtType] = del(self.chatEvents[evtType])
	end
	]]--
	ThreatLibNPCModuleCore.activeModule = nil
end

function ThreatLibNPCModuleCore.modulePrototype:CHAT_MSG_MONSTER_EMOTE(arg1, arg2)
	if not ThreatLib.running then
		return
	end
	local emote = self.chatEvents.emote
	if not emote then
		return
	end
	local mob = emote[arg2]
	if not mob then
		return
	end
	local func = mob[arg1]
	if not func then
		return
	end
	func(self)
end

function ThreatLibNPCModuleCore.modulePrototype:CHAT_MSG_MONSTER_YELL(arg1, arg2)
	if not ThreatLib.running then
		return
	end
	local yell = self.chatEvents.yell
	if not yell then
		return
	end

	-- Wowwiki says that arg1 = message, arg2 = monster
	-- Trust but verify!
	local mob = yell[arg2]
	if not mob then
		return
	end
	local func = mob[arg1]
	if not func then
		return
	end
	func(self)
end

function ThreatLibNPCModuleCore.modulePrototype:PLAYER_REGEN_ENABLED()
	self:ScheduleEvent("NPCModuleCoreResetMobCount" .. self.name, "ResetFight", 5)
end

function ThreatLibNPCModuleCore.modulePrototype:PLAYER_REGEN_DISABLED()
	if self:IsEventScheduled("NPCModuleCoreResetMobCount" .. self.name) then
		self:CancelScheduledEvent("NPCModuleCoreResetMobCount" .. self.name)
	end
end

function ThreatLibNPCModuleCore.modulePrototype:RegisterCombatant(combatantName, callback)
	combatantModuleLookup[combatantName] = self
	
	if not callback then
		return
	end
	if callback == true then
		callback = self.BossDeath
	end
	self.encounterEnemies[combatantName] = callback
	
	if useParser30 then
		self:RegisterParserEvent({
			eventType = "Damage",
			recipientName = combatantName
		}, "StartCombat")	

		self:RegisterParserEvent({
			eventType = "Cast",
			recipientName = combatantName
		}, "StartCombat")	
	else
		-- TODO: Parser-1.1 implementation
	end
end

-- Override to implement state reset on fight start
function ThreatLibNPCModuleCore.modulePrototype:StartFight()
end

function ThreatLibNPCModuleCore.modulePrototype:SetNumberOfMobs(num)
	self.encounterMobCount = max(1, num)
	ThreatLibNPCModuleCore.encounterMobCount = self.encounterMobCount
end

function ThreatLibNPCModuleCore.modulePrototype:GetNumberOfMobs()
	return self.encounterMobCount
end

function ThreatLibNPCModuleCore.modulePrototype:ResetFight()
	self:SetNumberOfMobs(1)
end

function ThreatLibNPCModuleCore.modulePrototype:StartCombat()
	if not ThreatLib.inCombat then
		self:StartFight()
	end
end

function ThreatLibNPCModuleCore.modulePrototype:BossDeath()
	self:ResetFight()
	ThreatLibNPCModuleCore:ToggleModuleActive(self, false)
end

local new, del = ThreatLib.new, ThreatLib.del

function ThreatLibNPCModuleCore.modulePrototype:RegisterChatEvent(eventType, mob, text, callback)
	if not mob then
		return
	end
	local t = self.chatEvents[eventType]
	if not t then
		t = new()
		self.chatEvents[eventType] = t
	end
	
	local tm = t[mob]
	if not tm then
		tm = new()
		t[mob] = tm
	end
	tm[text] = callback		
end

function ThreatLibNPCModuleCore.modulePrototype:UnregisterChatEvent(eventType, mob, text)
	local t = self.chatEvents[eventType]
	if not t then
		return
	end
	
	local tm = t[mob]
	if not tm then
		return
	end
	
	tm[text] = nil
	if not next(tm) then
		t[mob] = del(tm)
	end
	if not next(t) then
		self.chatEvents[eventType] = del(t)
	end
end

----------------------------------------------------------------
-- Module API
----------------------------------------------------------------

function ThreatLibNPCModuleCore.modulePrototype:WipeRaidThreatOnMob(mob)
	ThreatLib:WipeRaidThreatOnMob(mob)
end

table.insert(ThreatLib.UpvalueFixers, function(lib)
	ThreatLib = lib
	ThreatLibNPCModuleCore = ThreatLib:GetModule("NPCCore")
end)

end
